/**
 *
 * Copyright (C) 2013 Olivier Berthier
 *                    http://www.xunil.fr
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 **/

#ifndef BBOX_H
#define BBOX_H

#include <cassert>
#include "vector.h"
//boites englobantes pour définir des zones d'érosion


class BBox
{
public:
    // BBox Public Methods
    //! constructeur par defaut.
    BBox()
    {
        clear();
    }
    
    //! re-initialise les extremites.
    void clear( )
    {
        pMin = Vector( (float) HUGE_VAL,  (float) HUGE_VAL,  (float) HUGE_VAL );
        pMax = Vector( (float) -HUGE_VAL, (float) -HUGE_VAL, (float) -HUGE_VAL );
    }
    
    //! construction de l'aabbox d'un point, cf. Union( aabox, q ) pour ajouter d'autres points.
    BBox( const Vector &p ) 
        : 
        pMin( p ), pMax( p ) 
    { }
    
    //! construction d'une aabox connaissant 2 points.
    BBox( const Vector &p1, const Vector &p2 )
    {
        pMin = Vector( 
            std::min( p1[0], p2[0] ),
            std::min( p1[1], p2[1] ),
            std::min( p1[2], p2[2] ) );
        
        pMax = Vector( 
            std::max( p1[0], p2[0] ),
            std::max( p1[1], p2[1] ),
            std::max( p1[2], p2[2] ) );
    }
       
    void print( ) const
    {
        printf("[ %.10f %.10f %.10f ] x [ %.10f %.10f %.10f ] ", 
            pMin[0], pMin[1], pMin[2], 
            pMax[0], pMax[1], pMax[2]);
        
        Vector d( pMax-pMin);
        printf("extents (%.10f %.10f %.10f)\n",
            d[0], d[1], d[2]);
    }
    
    //! ajoute un point dans l'aabox.
    friend BBox Union( const BBox &b, const Vector &p );
    
    //! ajoute une autre aabox.
    friend BBox Union( const BBox &b, const BBox &b2 );
    
    //! ajoute une autre aabox
    void Union( const BBox& bbox )
    {
    #if 0
        pMin[0]= std::min(pMin[0], bbox.pMin[0]);
        pMin[1]= std::min(pMin[1], bbox.pMin[1]);
        pMin[2]= std::min(pMin[2], bbox.pMin[2]);
        pMax[0]= std::max(pMax[0], bbox.pMax[0]);
        pMax[1]= std::max(pMax[1], bbox.pMax[1]);
        pMax[2]= std::max(pMax[2], bbox.pMax[2]);        
    #else
        // plus rapide ... 
        if(bbox.pMin[0] < pMin[0])
            pMin[0]= bbox.pMin[0];
        if(bbox.pMin[1] < pMin[1])
            pMin[1]= bbox.pMin[1];
        if(bbox.pMin[2] < pMin[2])
            pMin[2]= bbox.pMin[2];
        
        if(bbox.pMax[0] > pMax[0])
            pMax[0]= bbox.pMax[0];
        if(bbox.pMax[1] > pMax[1])
            pMax[1]= bbox.pMax[1];
        if(bbox.pMax[2] > pMax[2])
            pMax[2]= bbox.pMax[2];
    #endif
    }
    
    //! ajoute un point a la bbox.
    void Union( const Vector& p )
    {
    #if 0
        pMin[0]= std::min(pMin[0], p[0]);
        pMin[1]= std::min(pMin[1], p[1]);
        pMin[2]= std::min(pMin[2], p[2]);
        pMax[0]= std::max(pMax[0], p[0]);
        pMax[1]= std::max(pMax[1], p[1]);
        pMax[2]= std::max(pMax[2], p[2]);
    #else
        // plus rapide ... 
        if(p[0] < pMin[0])
            pMin[0]= p[0];
        if(p[1] < pMin[1])
            pMin[1]= p[1];
        if(p[2] < pMin[2])
            pMin[2]= p[2];
        
        if(p[0] > pMax[0])
            pMax[0]= p[0];
        if(p[1] > pMax[1])
            pMax[1]= p[1];
        if(p[2] > pMax[2])
            pMax[2]= p[2];
    #endif
    }
    
    //! verifie l'intersection de 2 aabbox.
    bool Overlaps( const BBox &b ) const
    {
        bool x = ( pMax[0] >= b.pMin[0] ) && ( pMin[0] <= b.pMax[0] );
        bool y = ( pMax[1] >= b.pMin[1] ) && ( pMin[1] <= b.pMax[1] );
        bool z = ( pMax[2] >= b.pMin[2] ) && ( pMin[2] <= b.pMax[2] );
        return ( x && y && z );
    }
    
    void Intersection( const BBox& b )
    {
        pMin[0]= std::max(pMin[0], b.pMin[0]);
        pMin[1]= std::max(pMin[1], b.pMin[1]);
        pMin[2]= std::max(pMin[2], b.pMin[2]);
        pMax[0]= std::min(pMax[0], b.pMax[0]);
        pMax[1]= std::min(pMax[1], b.pMax[1]);
        pMax[2]= std::min(pMax[2], b.pMax[2]);        
    }
    
    friend BBox Intersection( const BBox& a, const BBox& b );
    
    bool isEmpty( ) const
    {
        bool x = ( pMax[0] < pMin[0] );
        bool y = ( pMax[1] < pMin[1] );
        bool z = ( pMax[2] < pMin[2] );
        return ( x && y && z );
    }
    
    //! verifie l'inclusion d'un point dans l'aabox.
    bool Inside( const Vector &pt ) const
    {
        return ( pt[0] >= pMin[0] && pt[0] <= pMax[0] 
            && pt[1] >= pMin[1] && pt[1] <= pMax[1] 
            && pt[2] >= pMin[2] && pt[2] <= pMax[2] );
    }
    
    //! dilate l'aabox de 'delta' dans toutes les directions.
    void Expand( float delta )
    {
        pMin -= Vector( delta, delta, delta );
        pMax += Vector( delta, delta, delta );
    }
    
    //! renvoie le centre de la boite.
    const Vector getCenter( ) const
    {
        return (pMin + pMax) * .5f;
    }
    
    //! renvoie le centre de la boite, sur un axe particulier.
    float getCenter( const int axis) const
    {
        return (pMin[axis] + pMax[axis]) * .5f;
    }
    
    //! renvoie le centre de la boite
    void getCenter( Vector& center ) const
    {
        center= getCenter();
    }
    
    //! renvoie le volume de l'aabox.
    float Volume( ) const
    {
        const Vector d( pMax-pMin);
        return d[0] * d[1] * d[2];
    }
    
    //! renvoie l'aire de l'aabox
    float SurfaceArea( ) const
    {
        const Vector d(pMax-pMin);
        const float area= 2.f * d[0] * d[1] + 2.f * d[0] * d[2] + 2.f * d[1] * d[2];
        return area;
    }
    
    //! renvoie l'indice de l'axe le plus long de l'aabox, 0 pour X, 1 pour Y et 2 pour Z.
    int MaximumExtent( ) const
    {
        const Vector d( pMax-pMin);
        if ( d[0] > d[1] && d[0] > d[2] )
            return 0;
        if ( d[1] > d[2] )
            return 1;
        else
            return 2;
    }
    
    //! renvoie une sphere englobante de l'aabox.
    void BoundingSphere( Vector &c, float &r ) const;
  
  
    //! acces indexe aux extremes de l'aabox.
    Vector& operator[] ( const int id )
    {
        assert(id == 0 || id == 1);
        return (&pMin)[id];
    }

    //! acces indexe aux extremes de l'aabox.
    const Vector& operator[] ( const int id ) const
    {
        assert(id == 0 || id == 1);
        return (&pMin)[id];
    }
    
    // BBox Public Data
    //! points extremes de l'aabox.
    Vector pMin, pMax;
};




#endif
