/**
 *
 * Copyright (C) 2013 Olivier Berthier
 *                    http://www.xunil.fr
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 **/

// Blob
// Changelog 05.01.24

#ifndef __Blob__
#define __Blob__

#include "vector.h"
#include "box.h"

double frand(double fMin, double fMax);
int irand(int rang, int min=0);


class Blob;

// Generic Blob node
class BlobNode {
 protected:
  Box box;  //!< Bounding box of the node.
  Vector color;
 public:
  // BlobNode(const Vector& color=Vector(0,0,0)):color(color) {}
  BlobNode() {}
  virtual ~BlobNode() {}

  virtual Vector Color(const Vector&)=0;
  //! Compute the intensity of the field function at a given point in space.
  virtual double Intensity(const Vector&)=0;

  Box GetBox() const { return box; }

  int inside(const Vector&){ return 0;}
};

// Cubic blend
class Blend
{
 protected:
  double S;
  double R;
  
 public:
  Blend() {}
  Blend(const double&,const double&);
  virtual ~Blend() {}

  virtual double operator()(const double&) const;
};

// Vertex skeletal element
class BlobVertex:public BlobNode
{
 protected:
  Vector c;  //!< Center.
  Blend blend;
 public: 
  BlobVertex(const Vector&,const double&,const double&,const Vector& = Vector(0,0,0));
  virtual ~BlobVertex() {}

  virtual double Intensity(const Vector&);
  virtual Vector Color(const Vector&);
};



// Blending node
class BlobBlend:public BlobNode
{
 protected:
  BlobNode* elements[2]; //!< Set of blending elements.
 public:
  BlobBlend(BlobNode*,BlobNode*);
  virtual ~BlobBlend();
  virtual double Intensity(const Vector&);
  virtual Vector Color(const Vector&);
};

class BlobUnion:public BlobNode
{
 protected:
  BlobNode* elements[2]; //!< Set of blending elements.
 public:
  BlobUnion(BlobNode*,BlobNode*);
  virtual ~BlobUnion();
  virtual double Intensity(const Vector&);
  virtual Vector Color(const Vector&);
};


class BlobIntersect:public BlobNode
{
 protected:
  BlobNode* elements[2]; //!< Set of blending elements.
 public:
  BlobIntersect(BlobNode*,BlobNode*);
  virtual ~BlobIntersect();
  virtual double Intensity(const Vector&);
  virtual Vector Color(const Vector&);
};

class BlobSub:public BlobNode
{
 protected:
  BlobNode* elements[2]; //!< Set of blending elements.
 public:
  BlobSub(BlobNode*,BlobNode*);
  virtual ~BlobSub();
  virtual double Intensity(const Vector&);
  virtual Vector Color(const Vector&);
};

class BlobXor:public BlobNode
{
 protected:
  BlobNode* elements[2]; //!< Set of blending elements.
 public:
  BlobXor(BlobNode*,BlobNode*);
  virtual ~BlobXor();
  virtual double Intensity(const Vector&);
  virtual Vector Color(const Vector&);
};

// Blobby shapes
class Blob
{
 protected:
  double threshold;  //!< Threshold value.
  BlobNode* element;//!< Root node.
 public:
  Blob(BlobNode* =NULL,const double& =0.5);

  virtual ~Blob();

  //! Return the threshold value.
  double Threshold() { return threshold; }

  double Intensity(const Vector&) const;
  Vector Gradient(const Vector&) const;

  Vector Dichotomy(Vector a,Vector b,double va,double vb,double length,const double& epsilon) const;
  Vector Color(const Vector&);
  void Polygonize(Box box,int n,Vector* vertex,int* triangles,Vector* color,int& nv,int& nt,const double& epsilon);

  BlobNode* Erode(const Box& box, const unsigned int& size,const double& epsilon,const Vector& color=Vector(0,0,0), const double& min=0.0, const double& max=1.0, const double& imin=0.5, const double& imax=1.5, const Vector& orientation=Vector(0,0,0),const Box& boxLocal=Box(Vector(0,0,0)));

  Vector searchPoint(const Box& box, const Vector& o, const double& epsilon,const double& epsilon2, const double& min, const double& max, const double& imin, const double& imax,const Box& boxLocal=Box(Vector(0,0,0)) );

  Box GetBox() const { return element->GetBox(); }
 protected:
  static int triTable[256][16];
};


class Quad {
 protected:
  Vector c;
  Vector n;
  Vector u;
  Vector v;
  double r;
  int size;
  Vector *p;
  double *a;
 public:
  Quad(const Vector&,const Vector&,int,const double&,Blob*);
  ~Quad();
  void Render();
  inline int Index(int i,int j) const { return i*size+j; }
};
#endif
